from typing import Iterable, List, Optional, Tuple

from vien.exceptions import PyFileArgNotFoundExit


def list_left_partition(items: Iterable[str], split: str) \
        -> Tuple[List[str], List[str]]:
    # todo test
    left = list()
    right = list()

    found = False
    for arg in items:
        if not found:
            if arg == split:
                found = True
                continue
            left.append(arg)
        else:
            right.append(arg)

    return left, right


class ParsedCall:
    __slots__ = ['args', 'file', 'file_idx']

    def __init__(self, args: List[str]):

        self.args = args

        call_found = False
        file_found = False
        for idx, arg in enumerate(args):
            if not call_found:
                if arg == "call":
                    call_found = True
                continue

            if arg.lower().endswith(".py"):
                self.file = arg
                self.file_idx = idx
                file_found = True
                break

        if not file_found:
            raise PyFileArgNotFoundExit

    @property
    def before_filename(self) -> Optional[str]:
        if self.file_idx <= 0:
            return None
        val = self.args[self.file_idx - 1]
        if val == "call":
            return None
        return val

# def call_pyfile(args: List[str]) -> Optional[str]:
#     # todo remove?
#     for arg in items_after(args, "call"):
#         if arg.lower().endswith(".py"):
#             return arg
#     return None



# SPDX-FileCopyrightText: (c) 2020 Artëm IG <github.com/rtmigo>
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations

import json
import os
import shutil
import subprocess
import sys
import shlex
from pathlib import Path
from typing import *

from vien import is_posix
from vien._common import need_posix, is_windows, need_windows
from vien.arg_parser import Commands, Parsed
from vien.bash_runner import run_as_bash_script
from vien.call_parser import ParsedCall, list_left_partition
from vien.colors import Colors
from vien.escaping_cmd import cmd_escape_arg
from vien.exceptions import ChildExit, VenvExistsExit, VenvDoesNotExistExit, \
    PyFileNotFoundExit, PyFileArgNotFoundExit, FailedToCreateVenvExit, \
    FailedToClearVenvExit, CannotFindExecutableExit

verbose = False


def exe_name() -> str:
    return os.path.basename(sys.argv[0])


def get_vien_dir() -> Path:
    path_from_var = os.environ.get("VIENDIR")
    if path_from_var:
        path_from_var = os.path.expandvars(path_from_var)
        path_from_var = os.path.expanduser(path_from_var)
        return Path(path_from_var)
    else:
        # It looks like storing dot files in the home directory
        # is the de facto standard for both worlds.
        #
        # App    | POSIX     | Windows
        # -------|-----------|----------------------
        # VSCode | ~/.vscode | %USERPROFILE%\.vscode
        # AWS:   | ~/.aws    | %USERPROFILE%\.aws
        #
        return Path.home() / ".vien"


def run_bash_sequence(commands: List[str], env: Optional[Dict] = None) -> int:
    need_posix()

    # command || exit /b 666

    lines = [
        # shebang not necessary as we specify executable in subprocess.call
        "set -e",  # fail on first error
    ]

    lines.extend(commands)

    # Ubuntu really needs executable='/bin/bash'.
    # Otherwise the command is executed in /bin/sh, ignoring the hashbang,
    # but SH fails to execute commands like 'source'

    return subprocess.call("\n".join(lines),
                           shell=True,
                           executable='/bin/bash',
                           env=env)


def run_cmdexe_sequence(commands: List[str], env: Optional[Dict] = None) -> int:
    # todo test independently
    # This function does not work "officially" yet.

    need_windows()

    # raise NotImplemented

    # https://stackoverflow.com/questions/734598/how-do-i-make-a-batch-file-terminate-upon-encountering-an-error

    # unlike bash, cmd.exe returns exit code zero even if last command returned
    # non-zero. There is also no evident way to turn on 'set -e' mode, i.e.
    # exit on the first failure.
    #
    # We'll just glue all the commands with &&

    glued = " && ".join(f'( {c} )' for c in commands)

    # print(f"CMD running {glued}")

    return subprocess.call(glued,
                           shell=True,
                           # executable='/bin/bash',
                           env=env)


def venv_dir_to_python_exe(venv_dir: Path) -> Path:
    # this method is being tested indirectly each time the venv is created:
    # vien prints the path to executable after running this function

    if is_posix:
        parent = venv_dir / "bin"
        basenames = "python", "python3"
    else:
        parent = venv_dir / "Scripts"
        basenames = "python.exe", "python3.exe"

    for name in basenames:
        executable = parent / name
        if executable.exists():
            return executable

    raise Exception(f"Cannot find the Python interpreter in {venv_dir}.")


def windows_cmdexe_activate(venv_dir: Path) -> Path:
    # https://docs.python.org/3/library/venv.html
    assert is_windows
    return venv_dir / 'Scripts' / 'activate.bat'


def posix_bash_activate(venv_dir: Path) -> Path:
    # https://docs.python.org/3/library/venv.html
    assert is_posix
    return venv_dir / 'bin' / 'activate'


def arg_to_python_interpreter(argument: Optional[str]) -> str:
    if argument is None:
        return sys.executable
    exe = shutil.which(argument)
    if not exe:
        raise CannotFindExecutableExit(argument)
    return exe


def main_create(venv_dir: Path, interpreter: Optional[str]):
    if venv_dir.exists():
        raise VenvExistsExit("Virtualenv already exists.")

    exe = arg_to_python_interpreter(interpreter)

    print(f"Creating {venv_dir}")

    result = subprocess.run([exe, "-m", "venv", str(venv_dir)])
    if result.returncode == 0:
        print()
        print("The Python executable:")
        print(str(venv_dir_to_python_exe(venv_dir)))
    else:
        raise FailedToCreateVenvExit(venv_dir)


def main_delete(venv_dir: Path):
    if "_venv" not in venv_dir.name:
        raise ValueError(venv_dir)
    if not venv_dir.exists():
        raise VenvDoesNotExistExit(venv_dir)

    # todo try to use the same executable that created the environment
    # If we use sys.executable, we may clear the venv in some incompatible way.
    # But we can't just use executable from the venv when clearing it:
    # Windows will fail with [WinError 5] Access is denied: '...python.exe'

    # todo check we are not running the same executable we about to delete
    # python_exe = venv_dir_to_python_exe(venv_dir)
    print(f"Clearing {venv_dir}")

    result = subprocess.run(
        [sys.executable, "-m", "venv", "--clear", str(venv_dir)],
        capture_output=True, encoding=sys.stdout.encoding)
    if result.returncode != 0:
        print(f"stdout: {result.stdout}")
        print(f"stderr: {result.stderr}")
        raise FailedToClearVenvExit(venv_dir)
    print(f"Deleting {venv_dir}")
    shutil.rmtree(str(venv_dir))


def main_recreate(venv_dir: Path, interpreter: Optional[str]):
    if venv_dir.exists():
        main_delete(venv_dir)
    main_create(venv_dir=venv_dir, interpreter=interpreter)


def guess_bash_ps1():
    """Returns the default BASH prompt."""

    # TL;DR PS1 is often inaccessible for child processes of BASH. It means,
    # for scripts too.
    #
    # AG 2021: PS1 is not an environment variable, but a local variable of
    # the shell [2019](https://stackoverflow.com/a/54999265). It seems to be
    # true for both MacOS 10.13 and Ubuntu 18.04.
    #
    # We can see PS1 by typing "echo $PS1" to the prompt, but ...
    #
    # 1) script.sh with `echo $PS1`        | prints nothing MacOS & Ubuntu
    #
    # 2) module.py with                    | prints Null MacOS & Ubuntu
    #    `print(os.environ.get("PS1")      |
    #
    # 3) `bash -i -c "echo $PS1"`          | seems to be OK in Ubuntu
    #     from command line                |
    #
    # 4) `zsh -i -c "echo $PS1"`           | looks like a normal prompt in OSX
    #     from command line                |
    #
    # In Ubuntu (3) returns the same prompt that in used by terminal by default.
    # Although if the user customized their PS1, no guarantees, that (3) will
    # return the updated value.
    #
    # For MacOS, the prompt printed by (3) in not the same as seen in terminal
    # app. It returns boring "bash-3.2" instead of expected "host:dir user$".
    #
    # (4) on MacOS seems to return the correct "host:dir user$", but it is in
    # ZSH format.

    # try to return $PS1 environment variable:
    env_var = os.environ.get("PS1")
    if env_var is not None:
        return env_var

    # for MacOS return predefined constant PS1
    import platform
    if platform.system() == "Darwin":
        return r"\h:\W \u\$"  # default for MacOS up to Catalina

    # hope for the best in other systems
    return subprocess.check_output(
        ['/bin/bash', '-i', '-c', 'echo $PS1']).decode().rstrip()


def main_shell(dirs: Dirs, input: Optional[str], input_delay: Optional[float]):
    dirs.venv_must_exist()

    activate_path_quoted = shlex.quote(
        str(dirs.venv_dir / "bin" / "activate"))

    old_ps1 = os.environ.get("PS1") or guess_bash_ps1()

    if not old_ps1:
        old_ps1 = r"\h:\W \u\$"  # default from MacOS

    color_start = Colors.YELLOW
    color_end = Colors.NOCOLOR

    venv_name = dirs.project_dir.name
    new_ps1 = f"{color_start}({venv_name}){color_end}:{old_ps1} "

    commands = [f'source {activate_path_quoted}']

    bashrc_file = Path(os.path.expanduser("~/.bashrc"))

    if bashrc_file.exists():
        # Ubuntu
        commands.append(
            f"exec bash --rcfile <(cat {json.dumps(str(bashrc_file))} "
            f"&& echo 'PS1={json.dumps(new_ps1)}')")
    else:
        # MacOS
        commands.append(f"PS1={json.dumps(new_ps1)} exec bash")

    # we will use [input] for testing: we will send a command to the stdin of
    # the interactive sub-shell and later check whether the command was
    # executed.
    #
    # We will also provide [input_delay] parameter. This allows the check
    # whether
    # the sub-shell was really interactive: did it wait for the input
    #
    # Surprisingly, the sub-shell will immediately close after executing the
    # command.  It seems it closes immediately after the subprocess.Popen
    # closes the stdin. So it will not wait for "exit". But it serves the
    # task well

    cp = run_as_bash_script("\n".join(commands),
                            input=input.encode() if input else None,
                            input_delay=input_delay,
                            env=child_env(dirs.project_dir))

    # the vien will return the same exit code as the shell returned
    raise ChildExit(cp.returncode)


def bash_args_to_str(args: List[str]) -> str:
    return ' '.join(shlex.quote(arg) for arg in args)


def cmdexe_args_to_str(args: List[str]) -> str:
    return ' '.join(cmd_escape_arg(arg) for arg in args)


def main_run(dirs: Dirs, command: List[str]):
    dirs.venv_must_exist()

    sequence: List[str] = list()

    if is_posix:
        activate_file = posix_bash_activate(dirs.venv_dir)
        sequence.append(f'source {shlex.quote(str(activate_file))}')
        sequence.append(bash_args_to_str(command))
        run_func = run_bash_sequence
    elif is_windows:
        activate_file = windows_cmdexe_activate(dirs.venv_dir)
        sequence.append(f'CALL {activate_file}"')
        sequence.append(cmdexe_args_to_str(command))
        run_func = run_cmdexe_sequence
    else:
        raise AssertionError("Unexpected OS")
    #    activate_file = (dirs.venv_dir / 'bin' / 'activate').absolute()
    if not activate_file.exists():
        raise FileNotFoundError(activate_file)

    exit_code = run_func(sequence, env=child_env(dirs.project_dir))
    raise ChildExit(exit_code)


class Dirs:
    def __init__(self, project_dir: Union[str, Path] = '.'):
        self.project_dir = Path(project_dir).absolute()
        self.venv_dir = get_vien_dir() / (self.project_dir.name + "_venv")
        if verbose:
            print(f"Proj dir: {self.project_dir}")
            print(f"Venv dir: {self.venv_dir}")

    def venv_must_exist(self) -> Dirs:
        if not self.venv_dir.exists():
            raise VenvDoesNotExistExit(self.venv_dir)
        return self


def _insert_into_pythonpath(insert_me: str) -> str:
    # https://docs.python.org/3/using/cmdline.html#envvar-PYTHONPATH
    # "The format is the same as the shell’s PATH: one or more directory
    # pathnames separated by os.pathsep (e.g. colons on Unix or semicolons
    # on Windows)"
    #
    ####
    #
    # $PYTHONPATH and sys.path are not the same.
    # - $PYTHONPATH is the variable to be used by all Python interpreters
    #   in the current shell environment
    # - sys.path is values for the current interpreter
    #
    # Passing values from own sys.path to other interpreter would be a mistake.
    # The current interpreter may be Python 3.7, and the other interpreter is
    # Python 3.9. Our sys.path leads to 3.7 system libraries, that are
    # unnecessary and even cause errors if added to $PYTHONPATH of 3.9.
    #
    ######
    # keeping it simple: not parsing the old string, just adding a prefix

    old = os.environ.get("PYTHONPATH", '')
    return f'{insert_me}{os.pathsep}{old}'


def child_env(proj_path: Path) -> Optional[Dict]:
    env: Optional[Dict]
    if proj_path != Path.cwd():
        return {
            **os.environ,
            'PYTHONPATH': _insert_into_pythonpath(str(proj_path))
        }
    else:
        return None


def relative_fn_to_module_name(filename: str) -> str:
    # todo test
    if not filename.lower().endswith('.py'):
        raise ValueError("The filename does not end with '.py'.")
    filename = filename[:-3]
    assert not os.path.isabs(filename)
    assert not filename.split()[0] == ".."
    return filename.replace(os.path.sep, '.')


def replace_arg(args: List[str], old: str, new: List[str]) -> List[str]:
    """Replaces first occurrence of `old` with a list of `new` items (zero or
    more items). Raises exception if `old` not found.
    """

    # todo remove?
    result = list()
    replaced = False
    for arg in args:
        if not replaced and arg == old:
            result.extend(new)
            replaced = True
        else:
            result.append(arg)

    assert replaced
    return result


def relative_inner_path(child: Union[str, Path],
                        parent: Union[str, Path]) -> str:
    """(/abc/parent/xyz/child, /abc/parent) -> xyz/child
    Not only returns the "relative" path, but also checks
    it is really relative.
    """
    # todo unit test
    rel_path = os.path.relpath(child, parent)
    if rel_path.split()[0] == ".." or os.path.isabs(rel_path):
        raise ValueError(f"The {child} is not a child of {parent}.")
    return rel_path


def main_call(parsed: Parsed, dirs: Dirs):
    dirs.venv_must_exist()

    parsed_call = ParsedCall(parsed.args)
    assert parsed_call.file is not None

    if not os.path.exists(parsed_call.file):
        raise PyFileNotFoundExit(Path(parsed_call.file))

    if parsed_call.before_filename == "-m":
        # todo unit test
        # /abc/project/package/module.py -> package/module.py
        relative = relative_inner_path(parsed_call.file, dirs.project_dir)
        # package/module.py -> package.module
        module_name = relative_fn_to_module_name(relative)
        # replacing the filename in args with the module name.
        # It is already prefixed with -m
        args = parsed_call.args.copy()
        args[parsed_call.file_idx] = module_name
        # args to python are those after 'call' word
        _, args_to_python = list_left_partition(args, 'call')
        assert '-m' in args_to_python
        assert module_name in args_to_python
    else:
        args_to_python = parsed.args_to_python

    python_exe = venv_dir_to_python_exe(dirs.venv_dir)

    assert len(args_to_python) > 0
    args = [str(python_exe)] + args_to_python

    cp = subprocess.run(args, env=child_env(dirs.project_dir))

    raise ChildExit(cp.returncode)


def normalize_path(reference: Path, path: Path) -> Path:
    # todo test
    if path.is_absolute():
        return path
    return Path(os.path.normpath(reference / path))


def get_project_dir(parsed: Parsed) -> Path:
    if parsed.project_dir_arg is not None:
        if parsed.command == Commands.call:
            # for the 'call' the reference dir is the parent or .py file
            pyfile = ParsedCall(parsed.args).file
            if pyfile is None:
                raise PyFileArgNotFoundExit
            reference_dir = Path(pyfile).parent.absolute()
        else:
            # for other commands the reference dir is cwd
            reference_dir = Path(".").absolute()
        project_dir = normalize_path(reference_dir,
                                     Path(parsed.project_dir_arg))
    else:
        # when the --project-dir is not specified, the project dir is cwd
        assert parsed.project_dir_arg is None
        project_dir = Path('.')
    project_dir = project_dir.absolute()
    return project_dir


def main_entry_point(args: Optional[List[str]] = None):
    parsed = Parsed(args)

    # todo replace private _ns attrs with public properties

    dirs = Dirs(project_dir=get_project_dir(parsed))

    if parsed.command == Commands.create:
        main_create(dirs.venv_dir, parsed.python_executable)
    elif parsed.command == Commands.recreate:
        main_recreate(dirs.venv_dir,
                      parsed.python_executable)  # todo .existing()?
    elif parsed.command == Commands.delete:  # todo move 'existing' check from func?
        main_delete(dirs.venv_dir)
    elif parsed.command == Commands.path:
        print(dirs.venv_dir)  # does not need to be existing
    elif parsed.command == Commands.run:
        # todo allow running commands from strings
        main_run(dirs.venv_must_exist(), parsed.run_args)
    elif parsed.command == Commands.call:

        main_call(parsed, dirs)

    elif parsed.command == Commands.shell:
        main_shell(dirs, parsed.shell_input, parsed.shell_delay)
    else:
        raise ValueError



# SPDX-FileCopyrightText: (c) 2021 Artëm IG <github.com/rtmigo>
# SPDX-License-Identifier: BSD-3-Clause


import unittest


def color_escape(s: str):
    esc_open = r"\[\e[;"  # r"\e[" is not enough! https://superuser.com/a/367280
    # esc_open = r"\[\e[;"  # r"\e[" is not enough! https://superuser.com/a/367280
    esc_close = r"\]"

    return f"{esc_open}{s}{esc_close}"


class TestColorEscape(unittest.TestCase):
    def test(self):
        # it's easy to lose significant backslashes so
        self.assertEqual(color_escape("inner"), r"\[\e[;inner\]")


class Colors:
    GREEN = color_escape("32m")
    MAGENTA = color_escape("35m")
    YELLOW = color_escape("33m")
    CYAN = color_escape("36m")
    BLUE = color_escape("34m")
    NOCOLOR = color_escape("0m")



# Code taken from https://stackoverflow.com/a/29215357
# SPDX-FileCopyrightText: Holger Just
# SPDX-License-Identifier: CC BY-SA 3.0


import re


# original link on SO is broken:
#   http://blogs.msdn.com/b/twistylittlepassagesallalike/archive/2011/04/23/
#   everyone-quotes-arguments-the-wrong-way.aspx
# same link on web-archive:
#   https://web.archive.org/web/20161228144344/https://blogs.msdn.microsoft.com/
#   twistylittlepassagesallalike/2011/04/23/
#   everyone-quotes-command-line-arguments-the-wrong-way/


def cmd_escape_arg(arg):
    # Escape the argument for the cmd.exe shell.
    #
    # First we escape the quote chars to produce a argument suitable for
    # CommandLineToArgvW. We don't need to do this for simple arguments.

    if not arg or re.search(r'(["\s])', arg):
        arg = '"' + arg.replace('"', r'\"') + '"'

    return _escape_for_cmd_exe(arg)


def _escape_for_cmd_exe(arg):
    # Escape an argument string to be suitable to be passed to
    # cmd.exe on Windows
    #
    # This method takes an argument that is expected to already be properly
    # escaped for the receiving program to be properly parsed. This argument
    # will be further escaped to pass the interpolation performed by cmd.exe
    # unchanged.
    #
    # Any meta-characters will be escaped, removing the ability to e.g. use
    # redirects or variables.
    #
    # @param arg [String] a single command line argument to escape for cmd.exe
    # @return [String] an escaped string suitable to be passed as a program
    #   argument to cmd.exe

    meta_chars = '()%!^"<>&|'
    meta_re = re.compile(
        '(' + '|'.join(re.escape(char) for char in list(meta_chars)) + ')')
    meta_map = {char: "^%s" % char for char in meta_chars}

    def escape_meta_chars(m):
        char = m.group(1)
        return meta_map[char]

    return meta_re.sub(escape_meta_chars, arg)



# SPDX-FileCopyrightText: (c) 2021 Artëm IG <github.com/rtmigo>
# SPDX-License-Identifier: BSD-3-Clause

import subprocess
import time
from subprocess import Popen, TimeoutExpired, CalledProcessError, \
    CompletedProcess, PIPE

from vien._common import need_posix


def run_as_bash_script(script: str, timeout: float = None,
                       input_delay: float = None,
                       capture_output: bool = False,
                       input: bytes = None,
                       **kwargs
                       ) -> subprocess.CompletedProcess:
    """Runs the provided string as a .sh script."""

    need_posix()

    # we need executable='/bin/bash' for Ubuntu 18.04, it will run '/bin/sh'
    # otherwise. For MacOS 10.13 it seems to be optional
    return _run_with_input_delay(script, shell=True, executable='/bin/bash',
                                 timeout=timeout,
                                 input=input,
                                 capture_output=capture_output,
                                 input_delay=input_delay,
                                 **kwargs)


def _run_with_input_delay(*popenargs, input_delay: float = None,
                          input=None, timeout: float = None,
                          check: bool = False,
                          capture_output: bool = False,
                          **kwargs):
    """Basically the same as subprocess.run, but accepts input_delay
    parameter."""

    # This is almost an exact copy of subprocess.run (as of 2021-03-06).
    # SPDX-FileCopyrightText: 2003-2005 Peter Astrand <astrand@lysator.liu.se>
    # https://github.com/python/cpython/blob/master/Lib/subprocess.py

    # START of insert from another portion of subprocess.py
    try:
        import msvcrt
        import _winapi
        _mswindows = True
    except ModuleNotFoundError:
        _mswindows = False
    # END of insert

    if input is not None:
        if kwargs.get('stdin') is not None:
            raise ValueError('stdin and input arguments may not both be used.')
        kwargs['stdin'] = PIPE

    if capture_output:
        if kwargs.get('stdout') is not None or kwargs.get('stderr') is not None:
            raise ValueError('stdout and stderr arguments may not be used '
                             'with capture_output.')
        kwargs['stdout'] = PIPE
        kwargs['stderr'] = PIPE

    with Popen(*popenargs, **kwargs) as process:
        try:
            # START of modified code
            if input is not None and input_delay:
                time.sleep(input_delay)
            # END of modified code

            stdout, stderr = process.communicate(input, timeout=timeout)
        except TimeoutExpired as exc:
            process.kill()
            if _mswindows:
                # Windows accumulates the output in a single blocking
                # read() call run on child threads, with the timeout
                # being done in a join() on those threads.  communicate()
                # _after_ kill() is required to collect that and add it
                # to the exception.
                exc.stdout, exc.stderr = process.communicate()
            else:
                # POSIX _communicate already populated the output so
                # far into the TimeoutExpired exception.
                process.wait()
            raise
        except:  # Including KeyboardInterrupt, communicate handled that.
            process.kill()
            # We don't call process.wait() as .__exit__ does that for us.
            raise

        exit_code = process.poll()

        # START of modified code
        if exit_code is None:
            raise RuntimeError("The process was not terminated.")
        # END of modified code

        if check and exit_code:
            raise CalledProcessError(exit_code, process.args,
                                     output=stdout, stderr=stderr)

    return CompletedProcess(process.args, exit_code, stdout, stderr)



import argparse
import sys
from enum import Enum
from typing import List, Optional, Iterable

from vien._common import is_windows

from vien import is_posix

import vien
#from vien.call_parser import items_after


def version_message() -> str:
    return "\n".join([
        f"VIEN: Python Virtual Environments Tool {vien.__version__}",
        vien.__copyright__
    ])


def usage_doc():
    text = f"""{version_message()}

See a detailed intro at
https://github.com/rtmigo/vien_py#readme

"""

    doc = text.strip()
    above_first_line = ("-" * len(doc.splitlines()[0]))
    return f"{above_first_line}\n{doc}\n"


def items_after(items: Iterable[str], x: str) -> Iterable[str]:
    found = False
    for arg in items:
        if found:
            yield arg
        elif arg == x:
            found = True
    if not found:
        raise LookupError


def remove_leading_p(args: List[str]) -> List[str]:
    # fixing a problem that is outdated since 2021-05
    if len(args) < 2:
        return args
    if args[0] == "-p" or args[0] == "--project-dir":
        args = args[2:]
    return args


class Commands(Enum):
    create = "create"
    delete = "delete"
    recreate = "recreate"
    shell = "shell"
    run = "run"
    call = "call"
    path = "path"


class Parsed:
    PARAM_WINDOWS_ALL_ARGS = "--vien-secret-windows-all-args"

    def __init__(self, args: Optional[List[str]]):
        super().__init__()

        if args is None:
            args = sys.argv[1:]

        # secret parameter PARAM_WINDOWS_ALL_ARGS allows to run commands that
        # are not yet fully supported on Windows.
        enable_windows_all_args = self.PARAM_WINDOWS_ALL_ARGS in args
        if enable_windows_all_args:
            # for more transparent testing, I don't want this param
            # to ever affect posix behavior
            assert is_windows

        parser = argparse.ArgumentParser()

        parser.add_argument("--project-dir", "-p", default=None, type=str)

        # the following parameter is added only to avoid parsing errors.
        # Actually we use its value from `args` before running ArgumentParser
        parser.add_argument(self.PARAM_WINDOWS_ALL_ARGS, action='store_true',
                            help=argparse.SUPPRESS)

        subparsers = parser.add_subparsers(dest='command', required=True)

        parser_init = subparsers.add_parser(Commands.create.name,
                                            help="create new virtualenv")
        parser_init.add_argument('python', type=str, default=None,
                                 nargs='?')

        subparsers.add_parser(Commands.delete.name,
                              help="delete existing virtualenv")

        parser_reinit = subparsers.add_parser(
            Commands.recreate.name,
            help="delete existing virtualenv and create new")
        parser_reinit.add_argument('python', type=str, default=None,
                                   nargs='?')

        if is_posix or enable_windows_all_args:
            shell_parser = subparsers.add_parser(
                Commands.shell.name,
                help="dive into Bash sub-shell using the virtualenv")
            shell_parser.add_argument("--input", type=str, default=None)
            shell_parser.add_argument("--delay", type=float, default=None,
                                      help=argparse.SUPPRESS)

        if is_posix or enable_windows_all_args:
            parser_run = subparsers.add_parser(
                Commands.run.name,
                help="run a command inside the virtualenv")
            parser_run.add_argument('otherargs', nargs=argparse.REMAINDER)

        parser_call = subparsers.add_parser(
            Commands.call.name,
            help="run a script inside the virtualenv")
        # todo Remove it later. [call -p] is outdated since 2021-05
        parser_call.add_argument("--project-dir", "-p", default=None, type=str,
                                 dest="outdated_call_project_dir",
                                 help=argparse.SUPPRESS)
        # this arg is for help only. Actually it's buggy (at least in 3.7),
        # so we will never use its result, and get those args other way
        parser_call.add_argument('args_to_python', nargs=argparse.REMAINDER)

        subparsers.add_parser(
            Commands.path.name,
            help="show the supposed path of the virtualenv "
                 "for the current directory")

        if not args:
            print(usage_doc())
            parser.print_help()
            exit(2)

        self.args = args

        # it seems, nargs.REMAINDER is buggy in 2021:
        # https://bugs.python.org/issue17050
        #
        # For example, when the first REMAINDER argument is an option
        # such as "-d", argparse shows error instead of just remembering "-d"
        #
        # But "-d" actually can be the first REMAINDER arg after the CALL
        # command.
        #
        # That's why we parse args twice. First time with `parse_known_args` -
        # to get the command name. And then, if it's not CALL - we parse
        # again with a stricter parse_args.

        self._ns: argparse.Namespace
        unknown: List[str]

        self._ns, unknown = parser.parse_known_args(self.args)
        if self.command == Commands.call:
            self.args_to_python = list(items_after(args, 'call'))

            # if some of the unknown args are NOT after the 'call',
            # then we're failed to interpret the command. Parsing
            # it stricter
            bad_unrecognized = [unk for unk in unknown if
                                unk not in self.args_to_python]
            if bad_unrecognized:
                parser.error(f"unrecognized arguments: {bad_unrecognized}")
                raise AssertionError("Not expected to run this line")

            # todo Remove later. [call -p] is outdated since 2021-05
            self.args_to_python = remove_leading_p(self.args_to_python)
        else:
            # if some args were not recognized, parsing everything stricter
            if unknown:
                self._ns = parser.parse_args(self.args)

    @property
    def command(self) -> Commands:
        return Commands(self._ns.command)

    @property
    def project_dir_arg(self) -> Optional[str]:
        """Returns either outdated [call -p ARG] or normal [vien -p ARG]
        value."""
        project_dir_after_call = self._ns.__dict__ \
            .get('outdated_call_project_dir')
        if project_dir_after_call is not None:
            print("'vien call -p proj/dir file.py' syntax is outdated. "
                  "Use 'vien -p proj/dir call file.py'.")
            return project_dir_after_call
        return self._ns.project_dir

    @property
    def python_executable(self) -> Optional[str]:
        if self.command not in (Commands.create, Commands.recreate):
            raise RuntimeError
        # assert self._ns.python is not None
        return self._ns.python

    @property
    def shell_input(self) -> Optional[str]:
        if self.command != Commands.shell:
            raise RuntimeError
        return self._ns.input

    @property
    def shell_delay(self) -> Optional[float]:
        if self.command != Commands.shell:
            raise RuntimeError
        return self._ns.delay

    @property
    def run_args(self) -> List[str]:
        if self.command != Commands.run:
            raise RuntimeError
        return self._ns.otherargs



from pathlib import Path


# from vien.main import exe_name




class VienExit(SystemExit):
    """Base class for all the expected exceptions,
    that show the error message and stop the program."""

    def __init__(self, arg):
        super().__init__(arg)


class ChildExit(VienExit):
    """When a child process finished, and vien must exit too with
    the same code"""

    def __init__(self, exit_code: int):
        super().__init__(exit_code)


class VenvExistsExit(VienExit):  # todo does it return error code?
    pass


class VenvDoesNotExistExit(VienExit):
    def __init__(self, path: Path):
        super().__init__(f'Virtual environment "{path}" does not exist.\n'
                         f'You can create it with "create" command.')


class PyFileNotFoundExit(VienExit):
    def __init__(self, path: Path):
        super().__init__(f"File {path} not found.")


class PyFileArgNotFoundExit(VienExit):
    def __init__(self):
        super().__init__(f"The arguments to the 'call' command must "
                         f"include a .py file.")


class FailedToCreateVenvExit(VienExit):
    def __init__(self, path: Path):
        super().__init__(f"Failed to create virtualenv {path}.")


class FailedToClearVenvExit(VienExit):
    def __init__(self, path: Path):
        super().__init__(f"Failed to clear virtualenv {path}.")


class CannotFindExecutableExit(VienExit):
    def __init__(self, version: str):
        super().__init__(f"Cannot resolve '{version}' to an executable file.")



import os

is_windows = os.name == 'nt'
is_posix = os.name == 'posix'

if not is_windows and not is_posix:
    raise SystemExit(f"System '{os.name}' is not supported.")


class UnexpectedOsError(Exception):
    pass


def need_posix():
    if not is_posix:
        raise UnexpectedOsError


def need_windows():
    if not is_windows:
        raise UnexpectedOsError
